#!/bin/sh
#
# $Idaemons: /home/cvs/pkgtools/etc/pkgtools.sh,v 1.4 2002/01/22 08:05:05 knu Exp $

case "$1" in
start|restart)
    if [ -d /usr/local/lib/compat/pkg ]; then
	/sbin/ldconfig -m /usr/local/lib/compat/pkg
    fi
    ;;
stop)
    ;;
esac
